# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0


import re
from pathlib import Path
from typing import Any, Optional, Union

import msgspec
import ruamel.yaml
from msgspec.structs import replace

from . import models as cici_config

decoder = msgspec.json.Decoder(type=cici_config.File)
image_fqdn_regex = re.compile(r"^[\w\.-]+/")


def inject_variable_names(variables: dict[str, dict]) -> dict[str, dict]:
    # make sure each variable has its 'name' field set from its key
    patched = {}
    for key, value in variables.items():
        # if value is None or not a dict, treat it as empty dict
        if not isinstance(value, dict):
            value = {}
        # only add name if missing
        value.setdefault("name", key)
        patched[key] = value
    return patched


def patch_image(image: str, container_proxy: str = "${CONTAINER_PROXY}") -> str:
    """Patch in $CONTAINER_PROXY to image unless the following are true:

    A: Does the image URL contain ${CONTAINER_PROXY} (the literal string)
    B: Is the image URL a fully-qualified container URL?
    C: Does the image already start with a variable?
    """

    if not image:
        return image
    if container_proxy in image:
        return image

    if image_fqdn_regex.match(image):
        return image

    if image.startswith("$"):
        return image

    return f"{container_proxy}{image}"


def loads(
    text: str,
    gitlab_ci_jobs: Optional[dict[str, Any]] = None,
    precommit_hooks: Optional[dict[str, Any]] = None,
) -> cici_config.File:
    # parse YAML into fully-typed File object
    if gitlab_ci_jobs is None:
        gitlab_ci_jobs = {}
    if precommit_hooks is None:
        precommit_hooks = {}

    yaml = ruamel.yaml.YAML(typ="safe")
    data = yaml.load(text)
    # verify targets exists even if empty
    data.setdefault("targets", [])

    # Inject precommit/gitlab includes into each target
    for target in data["targets"]:
        if target["name"] in precommit_hooks:
            target["precommit_hook"] = {"name": target["name"]}
        if target["name"] in gitlab_ci_jobs:
            target["gitlab_include"] = {"name": target["name"]}

    if "variables" in data:
        data["variables"] = inject_variable_names(data["variables"])

    # decode into file_struct
    file_struct = decoder.decode(msgspec.json.encode(data))

    # post process to patch CONTAINER_PROXY to container images
    patched_targets = []
    for target in file_struct.targets:
        if target.container is not None:
            patched_container = replace(
                target.container,
                image=patch_image(target.container.image),
            )
            patched_targets.append(replace(target, container=patched_container))
        else:
            patched_targets.append(target)

    return replace(file_struct, targets=patched_targets)


def load(
    file: Union[str, Path],
    gitlab_ci_jobs: Optional[dict[str, Any]] = None,
    precommit_hooks: Optional[dict[str, Any]] = None,
) -> cici_config.File:
    return loads(
        open(file).read(),
        gitlab_ci_jobs=gitlab_ci_jobs,
        precommit_hooks=precommit_hooks,
    )



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import functools
from typing import Union

import cattrs
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override

from ...utils import (
    make_multiline_string,
    make_quoted_dict,
    make_quoted_list,
    make_quoted_list_or_string,
    make_quoted_string,
    make_scalar_list,
)
from . import models

CONVERTER = cattrs.Converter(omit_if_default=True, forbid_extra_keys=True)


make_dict_struct = functools.partial(
    make_dict_structure_fn,
    _cattrs_forbid_extra_keys=True,
)

make_dict_unstruct = functools.partial(
    make_dict_unstructure_fn,
    _cattrs_omit_if_default=True,
)


def quoted_list_struct_hook(object, __):
    return make_quoted_list(object)


def quoted_string_struct_hook(object, __):
    return make_quoted_string(object)


def multiline_string_struct_hook(object, __):
    return make_multiline_string(object)


def quoted_list_or_string_struct_hook(object, __):
    return make_quoted_list_or_string(object)


def quoted_dict_struct_hook(object, __):
    return make_quoted_dict(object)


def structure_image(object, __):
    if object is None:
        return None
    elif isinstance(object, str):
        return make_quoted_string(CONVERTER.structure(object, str))
    return CONVERTER.structure(object, models.Image)


def structure_include(object, __):
    if isinstance(object, str):
        return CONVERTER.structure(object, str)
    items = []
    for item in object:
        if "local" in item:
            items.append(CONVERTER.structure(item, models.IncludeLocal))
        elif "project" in item:
            items.append(CONVERTER.structure(item, models.IncludeProject))
        elif "remote" in item:
            items.append(CONVERTER.structure(item, models.IncludeRemote))
        elif "template" in item:
            items.append(CONVERTER.structure(item, models.IncludeTemplate))
    return items


def structure_script(object, __):
    if isinstance(object, str):
        return CONVERTER.structure(object, str)
    items = []
    for item in object:
        if isinstance(item, str):
            items.append(CONVERTER.structure(item, str))
        else:
            for entry in item:
                items.append(CONVERTER.structure(entry, str))
    return make_scalar_list(items)


def structure_service(object, __):
    if isinstance(object, str):
        return make_quoted_string(CONVERTER.structure(object, str))
    return CONVERTER.structure(object, models.Service)


def structure_variables(object, __):
    variables = {}
    for key, value in object.items():
        if isinstance(value, str):
            variables[key] = make_quoted_string(value)
        else:
            variables[key] = CONVERTER.structure(value, models.Variable)
    return variables


CONVERTER.register_structure_hook(
    Union[
        str,
        list[
            Union[
                models.IncludeLocal,
                models.IncludeProject,
                models.IncludeRemote,
                models.IncludeTemplate,
            ]
        ],
    ],
    structure_include,
)


CONVERTER.register_structure_hook(
    Union[str, list[str]], quoted_list_or_string_struct_hook
)

CONVERTER.register_structure_hook(
    Union[str, models.Service],
    lambda object, _: CONVERTER.structure(
        object, str if isinstance(object, str) else models.Service
    ),
)

CONVERTER.register_structure_hook(
    Union[str, models.CacheKey],
    lambda object, _: CONVERTER.structure(
        object, str if isinstance(object, str) else models.CacheKey
    ),
)

CONVERTER.register_structure_hook(
    Union[str, models.Environment],
    lambda object, _: CONVERTER.structure(
        object, str if isinstance(object, str) else models.Environment
    ),
)

CONVERTER.register_structure_hook(
    Union[int, models.Retry],
    lambda object, _: CONVERTER.structure(
        object, int if isinstance(object, int) else models.Retry
    ),
)

CONVERTER.register_structure_hook(
    Union[str, list[Union[str, list[str]]]],
    structure_script,
)

CONVERTER.register_structure_hook(
    models.Image,
    make_dict_struct(
        models.Image,
        CONVERTER,
        name=override(struct_hook=quoted_string_struct_hook),
        entrypoint=override(struct_hook=quoted_list_or_string_struct_hook),
    ),
)

CONVERTER.register_structure_hook(
    models.Service,
    make_dict_struct(
        models.Service,
        CONVERTER,
        name=override(struct_hook=quoted_string_struct_hook),
        command=override(struct_hook=quoted_list_or_string_struct_hook),
        entrypoint=override(struct_hook=quoted_list_or_string_struct_hook),
    ),
)

CONVERTER.register_structure_hook(
    models.Variable,
    make_dict_struct(
        models.Variable,
        CONVERTER,
        description=override(struct_hook=multiline_string_struct_hook),
        value=override(struct_hook=quoted_string_struct_hook),
        options=override(struct_hook=quoted_list_struct_hook),
    ),
)


def unstructure_script(cls):
    object = CONVERTER.unstructure(cls, str if isinstance(cls, str) else list[str])
    if isinstance(object, str):
        return object  # make_quoted_scalar_string(object)
    return make_scalar_list(object)


CONVERTER.register_unstructure_hook(
    Union[str, list[str]],
    unstructure_script,
)

CONVERTER.register_structure_hook(
    Union[str, models.Image, None],
    structure_image,
)

CONVERTER.register_structure_hook(
    models.Rule,
    make_dict_struct(
        models.Rule,
        CONVERTER,
        if_=override(rename="if"),
    ),
)
CONVERTER.register_unstructure_hook(
    models.Rule,
    make_dict_unstruct(
        models.Rule,
        CONVERTER,
        if_=override(rename="if"),
    ),
)


CONVERTER.register_structure_hook(
    models.Job,
    make_dict_struct(
        models.Job,
        CONVERTER,
        variables=override(struct_hook=structure_variables),
    ),
)

CONVERTER.register_structure_hook(
    models.File,
    make_dict_struct(
        models.File,
        CONVERTER,
        variables=override(struct_hook=structure_variables),
    ),
)
